//    DDS; 
//      .
//   STM32F100RB (24).
// : DAC  1 (PA4), DMA1  3, TIM6.

// F0  80    ;
//  400      .

#include <stm32f10x.h>

#include <math.h>

//   ,    
//     Vref/65536.
const int out_max=0xFFFF;
int out_a0=out_max/2;
int out_a1=out_max/3;

//      TIM6.
#define FSYSCLK 24000000
//     ( 
//    ).
//     FSYSCLK/K, 
// K-   2..65536.
//       ,
//    .
#define F0 80000

//    .
unsigned int ph_ind=0;

//       ;
//    f=F0*ph_m/(2.0**32),
//  F0 -  .
unsigned int ph_m=0;

//      .
#define Nt 8
//   .
const unsigned int L=1<<Nt;
//    .
const unsigned int tsize=L/4+1;

//  .
const unsigned int Nt_1=Nt-1;
const unsigned int Nt_2=Nt-2;
const unsigned int m=32-Nt;

// ,     
//    : (Nt-1)    .
const unsigned int ind_mask=((1<<Nt)-1)>>1;

//   sin   0..L/4 ( 0..M_PI/2);
//    , 15    .
unsigned short int table[tsize];    //fix15 format

//    .
void sin_table_init()
{
    for(unsigned int i=0; i<tsize; i++)
        table[i]=0x8000*sinf(2*(float)M_PI*i/L);
}

// ,   ,   sin 
//     : t=0..L-1
// (  t>=L,      L).
inline int select(unsigned int t)
{
    unsigned int res_sign=(t>>Nt_1)&1;
    unsigned int arg_sign=(t>>Nt_2)&1;
    t=(t^(-arg_sign))+arg_sign;
    t&=ind_mask;
    int r=table[t];
    return (r^(-res_sign))+res_sign;
}

// ,    sin,  
//    sin    
// (  -   32-   ).
inline int interp(unsigned int k)
{
    const unsigned int mask2=(1<<m)-1;
    const unsigned int m1=m-15;
    const int c=1<<15;

    unsigned int t1=k>>m;
    int y1=select(t1);
    int y2=select(t1+1);  //  t1==L-1 => t1+1==L .

    int dk=(k&mask2)>>m1;
    int r=(y1*(c-dk)+y2*dk)>>14;
    return r;
}

//  ,   DAC   DMA.
const unsigned int dma_buf_length=64;
short int dma_buf[dma_buf_length];

//    .
//    ph_ind,  ph_m.

//    .
inline void dma_buf_fragment_init(short int *begin, const short int *end)
{
    int a1=out_a1>>1; //    .
    while(begin!=end)
    {
        *begin=out_a0+
            (a1*interp(ph_ind)>>15); //   .
        ph_ind+=ph_m;
        begin++;
    }
}

//      .
inline void dma_buf_first_half_init()
{
    dma_buf_fragment_init(dma_buf, dma_buf+dma_buf_length/2);
}

//      .
inline void dma_buf_second_half_init()
{
    dma_buf_fragment_init(dma_buf+dma_buf_length/2, dma_buf+dma_buf_length);
}

//    DMA ( 3).
extern "C" void DMA1_Channel3_IRQHandler()
{
    if(DMA1->ISR&DMA_ISR_HTIF3)
        dma_buf_first_half_init();
    else if(DMA1->ISR&DMA_ISR_TCIF3)
        dma_buf_second_half_init();
    DMA1->IFCR|=                //  .   1.
            DMA_IFCR_CTEIF3|
            DMA_IFCR_CHTIF3|
            DMA_IFCR_CTCIF3|
            DMA_IFCR_CGIF3;
}

int main()
{
    //       DMA.
    NVIC_SetPriority(DMA1_Channel3_IRQn, 0);
    NVIC_EnableIRQ(DMA1_Channel3_IRQn);

    //     :
    // DMA, DAC, TIM6, GPIOA.
    RCC->AHBENR|=RCC_AHBENR_DMA1EN;
    RCC->APB1ENR|=RCC_APB1ENR_DACEN|RCC_APB1ENR_TIM6EN;
    RCC->APB2ENR|=RCC_APB2ENR_IOPAEN;

    // PA4 -       DAC.
    GPIOA->CRL&=~(0xf<<4*4);

    //   .
    sin_table_init();

    //       
    //  DAC  DMA.
    DMA1_Channel3->CPAR=(uint32_t)&DAC->DHR12L1; // 
    DMA1_Channel3->CMAR=(uint32_t)dma_buf;       //    .
    DMA1_Channel3->CNDTR=dma_buf_length;         // - .

    uint32_t tmp=DMA1_Channel3->CCR&~0x7FFF|
            DMA_CCR3_TCIE|      //    .
            DMA_CCR3_HTIE|      //   .
            DMA_CCR3_DIR|       // : ->.
            DMA_CCR3_CIRC|      //   .
            // DMA_CCR3_PINC=0  //    .
            DMA_CCR3_MINC|      //    .
            DMA_CCR3_PSIZE_1|   //   =10: 32 .
            DMA_CCR3_MSIZE_0|   //    =01: 16 .
            DMA_CCR3_PL;        // 11 -  .
    DMA1_Channel3->CCR=tmp;
    DMA1_Channel3->CCR|=DMA_CCR3_EN;

    //    DAC.
    DAC->CR|=
        DAC_CR_DMAEN1| //    DMA.
        DAC_CR_TEN1;   //   DAC    .
        // TSEL[2:0]=000:   Timer 6 TRGO event.
    DAC->CR|=DAC_CR_EN1;

    //       -
    //  TIM6.     
    //    f=F0*ph_m/(2.0**32) =>
    // ph_m=(uint32_t)((2.0**32)*f/F0);

    // :
    ph_m=2684354; // 50 
    //ph_m=2689723; // 50.1 
    //ph_m=268435456; // 5000 
    //ph_m=268435456*2; // 10000 
    //ph_m=662803111; // 12345.67Hz

    //        DAC,
    //         DMA.
    dma_buf_first_half_init();
    dma_buf_second_half_init();

    TIM6->ARR=FSYSCLK/F0-1;   //  : FSYSCLK/(ARR+1)=F0.
    TIM6->CR2&=~TIM_CR2_MMS;
    TIM6->CR2|=TIM_CR2_MMS_1; //MMS[2:0]=010:  TRGO  UEV.
    TIM6->CR1|=TIM_CR1_CEN;   //  .

    while(true);
}
